"""Code to help identify, extract, and file external downloads.

This package contains code to help you build importers and drive the
process of identifying which importer to run on an externally
downloaded file, extract transactions from them and file away these
files under a clean and rigidly named hierarchy for preservation.

"""
__copyright__ = "Copyright (C) 2016,2018  Martin Blais"
__license__ = "GNU GPLv2"


import os
import sys
import warnings
import click

from beancount import loader

from beangulp import archive
from beangulp import cache  # noqa: F401
from beangulp import exceptions
from beangulp import extract
from beangulp import identify
from beangulp import utils
from beangulp.importer import Importer, ImporterProtocol, Adapter


def _walk(file_or_dirs, log):
    """Convenience wrapper around beangulp.utils.walk()

    Log the name of the file being processed and check the input file
    size against identify.FILE_TOO_LARGE_THRESHOLD for too large input.

    """
    for filename in utils.walk(file_or_dirs):
        log(f'* {filename:}', nl=False)
        if os.path.getsize(filename) > identify.FILE_TOO_LARGE_THRESHOLD:
            log(' ... SKIP')
            continue
        yield filename


@click.command('extract')
@click.argument('src', nargs=-1, type=click.Path(exists=True, resolve_path=True))
@click.option('--output', '-o', type=click.File('w'), default='-',
              help='Output file.')
@click.option('--existing', '-e', type=click.Path(exists=True),
              help='Existing Beancount ledger for de-duplication.')
@click.option('--reverse', '-r', is_flag=True,
              help='Sort entries in reverse order.')
@click.option('--failfast', '-x', is_flag=True,
              help='Stop processing at the first error.')
@click.option('--quiet', '-q', count=True,
              help='Suppress all output.')
@click.pass_obj
def _extract(ctx, src, output, existing, reverse, failfast, quiet):
    """Extract transactions from documents.

    Walk the SRC list of files or directories and extract the ledger
    entries from each file identified by one of the configured
    importers.  The entries are written to the specified output file
    or to the standard output in Beancount ledger format in sections
    associated to the source document.

    """
    verbosity = -quiet
    log = utils.logger(verbosity, err=True)
    errors = exceptions.ExceptionsTrap(log)

    # Load the ledger, if one is specified.
    existing_entries = loader.load_file(existing)[0] if existing else []

    extracted = []
    for filename in _walk(src, log):
        with errors:
            importer = identify.identify(ctx.importers, filename)
            if not importer:
                log('') # Newline.
                continue

            # Signal processing of this document.
            log(' ...', nl=False)

            # Extract entries.
            entries = extract.extract_from_file(importer, filename, existing_entries)
            account = importer.account(filename)

            extracted.append((filename, entries, account, importer))
            log(' OK', fg='green')

        if failfast and errors:
            break

    # Sort.
    extract.sort_extracted_entries(extracted)

    # Deduplicate.
    for filename, entries, account, importer in extracted:
        importer.deduplicate(entries, existing_entries)
        existing_entries.extend(entries)

    # Invoke hooks.
    for func in ctx.hooks:
        extracted = func(extracted, existing_entries)

    # Serialize entries.
    extract.print_extracted_entries(extracted, output)

    if errors:
        sys.exit(1)


@click.command('archive')
@click.argument('src', nargs=-1, type=click.Path(exists=True, resolve_path=True))
@click.option('--destination', '-o', metavar='DIR',
              type=click.Path(exists=True, file_okay=False, resolve_path=True),
              help='The destination documents tree root directory.')
@click.option('--overwrite', '-f', is_flag=True,
              help='Overwrite destination files with the same name.')
@click.option('--dry-run', '-n', is_flag=True,
              help='Just print where the files would be moved.')
@click.option('--failfast', '-x', is_flag=True,
              help='Stop processing at the first error.')
@click.option('--quiet', '-q', count=True,
              help='Suppress all output.')
@click.pass_obj
def _archive(ctx, src, destination, dry_run, overwrite, failfast, quiet):
    """Archive documents.

    Walk the SRC list of files or directories and move each file
    identified by one of the configured importers in a directory
    hierarchy mirroring the structure of the accounts associated to
    the documents and with a file name composed by the document date
    and document name returned by the importer.

    Documents are moved to their filing location only when no errors
    are encountered processing all the input files.  Documents in the
    destination directory are not overwritten, unless the --force
    option is used.  When the directory hierarchy root is not
    specified with the --destination DIR options, it is assumed to be
    directory in which the ingest script is located.

    """
    # If the output directory is not specified, move the files at the
    # root where the import script is located. Providing this default
    # seems better than using a required option.
    if destination is None:
        import __main__
        destination = os.path.dirname(os.path.abspath(__main__.__file__))

    verbosity = -quiet
    log = utils.logger(verbosity, err=True)
    errors = exceptions.ExceptionsTrap(log)
    renames = []

    for filename in _walk(src, log):
        with errors:
            importer = identify.identify(ctx.importers, filename)
            if not importer:
                log('') # Newline.
                continue

            # Signal processing of this document.
            log(' ...', nl=False)

            destpath = archive.filepath(importer, filename)

            # Prepend destination directory path.
            destpath = os.path.join(destination, destpath)

            # Check for destination filename collisions.
            collisions = [src for src, dst in renames if dst == destpath]
            if collisions:
                raise exceptions.Error('Collision in destination file path.', destpath)

            # Check if the destination file already exists.
            if not overwrite and os.path.exists(destpath):
                raise exceptions.Error('Destination file already exists.', destpath)

            renames.append((filename, destpath))
            log(' OK', fg='green')
            log(f'  {destpath:}')

        if failfast and errors:
            break

    # If there are any errors, stop here.
    if errors:
        log('# Errors detected: documents will not be filed.')
        sys.exit(1)

    if not dry_run:
        for filename, destpath in renames:
            archive.move(filename, destpath)


@click.command('identify')
@click.argument('src', nargs=-1, type=click.Path(exists=True, resolve_path=True))
@click.option('--failfast', '-x', is_flag=True,
              help='Stop processing at the first error.')
@click.option('--verbose', '-v', is_flag=True,
              help='Show account information.')
@click.pass_obj
def _identify(ctx, src, failfast, verbose):
    """Identify files for import.

    Walk the SRC list of files or directories and report each file
    identified by one of the configured importers.  When verbose
    output is requested, also print the account name associated to the
    document by the importer.

    """
    log = utils.logger(verbose)
    errors = exceptions.ExceptionsTrap(log)

    for filename in _walk(src, log):
        with errors:
            importer = identify.identify(ctx.importers, filename)
            if not importer:
                log('') # Newline.
                continue

            # Signal processing of this document.
            log(' ...', nl=False)

            # When verbose output is requested, get the associated account.
            account = importer.account(filename) if verbose else None

            log(' OK', fg='green')
            log(f'  {importer.name:}')
            log(f'  {account:}', 1)

        if failfast and errors:
            break

    if errors:
        sys.exit(1)


def _importer(importer):
    """Check that the passed instance implements the Importer interface.

    Wrap ImporterProtocol instances with Adapter as needed.

    """
    if isinstance(importer, Importer):
        return importer
    if isinstance(importer, ImporterProtocol):
        warnings.warn('The beangulp.importer.ImporterProtocol interface for '
                      'importers has been replaced by the beangulp.Importer '
                      'interface and is therefore deprecated. Please update '
                      'your importer {} to the new interface.'.format(importer),
                      stacklevel=3)
        return Adapter(importer)
    raise TypeError(f'expected bengulp.Importer not {type(importer):}')


class Ingest:
    def __init__(self, importers, hooks=None):
        self.importers = [_importer(i) for i in importers]
        self.hooks = list(hooks) if hooks is not None else []

        while extract.find_duplicate_entries in self.hooks:
            self.hooks.remove(extract.find_duplicate_entries)
            warnings.warn('beangulp.extract.find_duplicate_entries has been removed '
                          'from the import hooks. Deduplication is now integral part '
                          'of the extract processing and can be customized by the '
                          'importers. See beangulp.importer.Importer.', stacklevel=2)

        @click.group('beangulp')
        @click.version_option()
        @click.pass_context
        def cli(ctx):
            """Import data from and file away documents from financial institutions."""
            ctx.obj = self

        cli.add_command(_archive)
        cli.add_command(_extract)
        cli.add_command(_identify)

        self.cli = cli

    def __call__(self):
        return self.cli()



"""Utilities using petl.
"""

from typing import Optional
import datetime
import re

import petl

from beancount.core import data
from beancount.core import amount
from beancount.core import flags


petl.config.look_style = "minimal"
petl.config.failonerror = True


def table_to_directives(
    table: petl.Table, currency: str = "USD", filename: Optional[str] = None
) -> data.Entries:
    """Convert a petl table to Beancount directives.

    This is intended as a convenience for many simple CSV importers. Your CSV
    code uses petl to normalize the contents of an imported file to a table, and
    this routine is called to actually translate that into directives.

    Required columns of the input 'table' are:
      date: A datetime.date instance, for the transaction.
      account: An account string, for the posting.
      amount: A Decimal instance, the number you want on the posting.
    Optional columns are:
      payee: A string, for the transaction's payee.
      narration: A string, for the transaction's narration.
      balance: A Decimal, the balance in the account *after* the given transaction.
      other_account: An account string, for the remainder of the transaction.
    """
    # Ensure the table is sorted in order to produce the final balance.
    assert table.issorted("date")
    assert set(table.fieldnames()) >= {"date", "account", "amount"}

    columns = table.fieldnames()
    metas = []
    for column in columns:
        match = re.match("meta:(.*)", column)
        if match:
            metas.append((column, match.group(1)))

    # Create transactions.
    entries = []
    filename = filename or f"<{__file__}>"
    for index, rec in enumerate(table.records()):
        meta = data.new_metadata(filename, index)
        units = amount.Amount(rec.amount, currency)
        tags, links = set(), set()
        txn = data.Transaction(
            meta,
            rec.date,
            flags.FLAG_OKAY,
            getattr(rec, "payee", None),
            getattr(rec, "narration", ""),
            tags,
            links,
            [data.Posting(rec.account, units, None, None, None, None)],
        )
        if hasattr(rec, "other_account") and rec.other_account:
            txn.postings.append(
                data.Posting(rec.other_account, None, None, None, None, None)
            )
        link = getattr(rec, "link", None)
        if link:
            links.add(link)
        tag = getattr(rec, "tag", None)
        if tag:
            tags.add(tag)

        for column, key in metas:
            value = getattr(rec, column, None)
            if value:
                meta[key] = value
        entries.append(txn)

    if "balance" in columns:
        # Insert a balance with the final value.
        meta = data.new_metadata(filename, index + 1)
        balance_date = rec.date + datetime.timedelta(days=1)
        entries.append(
            data.Balance(
                meta,
                balance_date,
                rec.account,
                amount.Amount(rec.balance, currency),
                None,
                None,
            )
        )

    return entries


def absorb_extra(table: petl.Table, column: str) -> petl.Table:
    """Absorb extra columns in a specific column.

    Sadly, some programmers in banks will forego the usage of libraries
    performing proper escaping of commas. This produces invalid CSV files, where
    some rows have an extra column or two. This function fixes up those mistakes
    by specifying a column to absort extra columns.

    The behavior is as follows: if the header has 7 columns, we assume the rest
    of the rows in the file should also have 7 columns. If this function is
    given the column name 'description' and it lives in the 5th column, a row
    with an abnormal 8 columns will be modified to merge the 5th nd 6th columns.
    If a row shows up with 9 columns, the 5th column would absort columns 5, 6
    and 7.  Rows with the normal 7 columns are unaffected.
    """
    header = table.fieldnames()
    num_expected_cols = len(header)
    absorbent_col = header.index(column)

    def absorb(row):
        row = list(row)
        if len(row) > num_expected_cols:
            for _ in range(len(row) - num_expected_cols):
                row[absorbent_col] = "{}, {}".format(
                    row[absorbent_col], row[absorbent_col + 1]
                )
                del row[absorbent_col + 1]
        return tuple(row)

    return table.rowmap(absorb, header=header)



"""CSV importer."""

# Postpone evaluation of annotations. This solves an issue with the
# import of the standard library csv module when this module or its
# associated test are executed as scripts. This requires Python >3.7
# and has to wait till we will drop suppport for Python 3.6
# from __future__ import annotations

__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

from inspect import signature
from os import path
from typing import Callable, Dict, List, Optional, Union
import collections
import csv
import datetime
import enum
import io
import warnings

import dateutil.parser

from beancount.core import data
from beancount.core import flags
from beancount.core.amount import Amount
from beancount.core.number import ZERO, D
from beangulp import utils
from beangulp import cache
from beangulp import date_utils
from beangulp import importer
from beangulp.importers.mixins import filing, identifier


class Col(enum.Enum):
    """The set of interpretable columns."""

    # The settlement date, the date we should create the posting at.
    DATE = '[DATE]'

    # The date at which the transaction took place.
    TXN_DATE = '[TXN_DATE]'

    # The time at which the transaction took place.
    # Beancount does not support time field -- just add it to metadata.
    TXN_TIME = '[TXN_TIME]'

    # The payee field.
    PAYEE = '[PAYEE]'

    # The narration fields. Use multiple fields to combine them together.
    NARRATION = NARRATION1 = '[NARRATION1]'
    NARRATION2 = '[NARRATION2]'
    NARRATION3 = '[NARRATION3]'

    # The amount being posted.
    AMOUNT = '[AMOUNT]'

    # Debits and credits being posted in separate, dedicated columns.
    AMOUNT_DEBIT = '[DEBIT]'
    AMOUNT_CREDIT = '[CREDIT]'

    # The balance amount, after the row has posted.
    BALANCE = '[BALANCE]'

    # A field to use as a tag name.
    TAG = '[TAG]'

    # A field to use as a unique reference id or number.
    REFERENCE_ID = '[REF]'

    # A column which says DEBIT or CREDIT (generally ignored).
    DRCR = '[DRCR]'

    # Last 4 digits of the card.
    LAST4 = '[LAST4]'

    # An account name.
    ACCOUNT = '[ACCOUNT]'

    # Categorization, if the institution supports it. You could, in theory,
    # specialize your importer to use this automatically assign a good expenses
    # account.
    CATEGORY = '[CATEGORY]'

    # A column that indicates the amount currency for the current row which may
    # be different to the base currency.
    CURRENCY = '[CURRENCY]'


def get_amounts(iconfig, row, allow_zero_amounts, parse_amount):
    """Get the amount columns of a row.

    Args:
      iconfig: A dict of Col to row index.
      row: A row array containing the values of the given row.
      allow_zero_amounts: Is a transaction with amount D('0.00') okay? If not,
        return (None, None).
    Returns:
      A pair of (debit-amount, credit-amount), both of which are either an
      instance of Decimal or None, or not available.
    """
    debit, credit = None, None
    if Col.AMOUNT in iconfig:
        credit = row[iconfig[Col.AMOUNT]]
    else:
        debit = row[iconfig[Col.AMOUNT_DEBIT]] if Col.AMOUNT_DEBIT in iconfig else None
        credit = row[iconfig[Col.AMOUNT_CREDIT]] if Col.AMOUNT_CREDIT in iconfig else None

    # If zero amounts aren't allowed, return null value.
    is_zero_amount = ((credit is not None and parse_amount(credit) == ZERO) and
                      (debit is not None and parse_amount(debit) == ZERO))
    if not allow_zero_amounts and is_zero_amount:
        return (None, None)

    return (-parse_amount(debit) if debit else None,
            parse_amount(credit) if credit else None)


def normalize_config(config, head, dialect='excel', skip_lines: int = 0):
    """Using the header line, convert the configuration field name lookups to int indexes.

    Args:
      config: A dict of Col types to string or indexes.
      head: A string, some decent number of bytes of the head of the file.
      dialect: A dialect definition to parse the header
      skip_lines: Skip first x (garbage) lines of file.
    Returns:
      A pair of
        A dict of Col types to integer indexes of the fields, and
        a boolean, true if the file has a header.
    Raises:
      ValueError: If there is no header and the configuration does not consist
        entirely of integer indexes.
    """
    # Skip garbage lines before sniffing the header
    assert isinstance(skip_lines, int)
    assert skip_lines >= 0

    head = io.StringIO(head, newline=None)
    lines = list(head)[skip_lines:]

    has_header = csv.Sniffer().has_header('\n'.join(lines))
    if has_header:
        header = next(csv.reader(lines, dialect=dialect))
        field_map = {name.strip(): index for index, name in enumerate(header)}
        index_config = {}
        for field_type, field in config.items():
            if isinstance(field, str):
                field = field_map[field]
            index_config[field_type] = field
    else:
        if any(not isinstance(field, int)
               for field_type, field in config.items()):
            raise ValueError("CSV config without header has non-index fields: "
                             "{}".format(config))
        index_config = config
    return index_config, has_header


def prepare_for_identifier(regexps: Union[str, List[str]],
                           matchers: Optional[List[str]]) -> Dict[str, str]:
    """Prepare data for identifier mixin."""
    if isinstance(regexps, str):
        regexps = [regexps]
    matchers = matchers or []
    matchers.append(('mime', 'text/csv'))
    if regexps:
        for regexp in regexps:
            matchers.append(('content', regexp))
    return {'matchers': matchers}


def prepare_for_filing(account: str, institution: Optional[str],
                       prefix: Optional[str]) -> Dict[str, str]:
    """Prepare kwds for filing mixin."""
    kwds = {'filing': account}
    if institution:
        prefix = kwds.get('prefix', None)
        assert prefix is None
        kwds['prefix'] = institution
    return kwds


class _CSVImporterBase:
    """Base class for CSV importer implementations.

    Note that many existing importers are based on this; be careful with
    modification of the attribute names and types. See concrete implementations
    below.
    """
    # pylint: disable=too-many-instance-attributes

    FLAG = flags.FLAG_OKAY

    def __init__(self, config, account, currency,
                 regexps=None,
                 skip_lines: int = 0,
                 last4_map: Optional[Dict] = None,
                 categorizer: Optional[Callable] = None,
                 institution: Optional[str] = None,
                 debug: bool = False,
                 csv_dialect: Union[str, csv.Dialect] = 'excel',
                 dateutil_kwds: Optional[Dict] = None,
                 narration_sep: str = '; ',
                 encoding: Optional[str] = None,
                 invert_sign: Optional[bool] = False,
                 **kwds):
        """Constructor.

        Args:
          config: A dict of Col enum types to the names or indexes of the columns.
          account: An account string, the account to post this to.
          currency: A currency string, the currency of this account.
          regexps: A list of regular expression strings.
          skip_lines: Skip first x (garbage) lines of file.
          last4_map: A dict that maps last 4 digits of the card to a friendly string.
          categorizer: A callable with two arguments (transaction, row) that can attach
            the other posting (usually expenses) to a transaction with only single posting.
          institution: An optional name of an institution to rename the files to.
          debug: Whether or not to print debug information
          csv_dialect: A `csv` dialect given either as string or as instance or
            subclass of `csv.Dialect`.
          dateutil_kwds: An optional dict defining the dateutil parser kwargs.
          narration_sep: A string, a separator to use for splitting up the payee and
            narration fields of a source field.
          encoding: Encoding for the file, utf-8 if not specified or None.
          invert_sign: If true, invert the amount's sign unconditionally.
          **kwds: Extra keyword arguments to provide to the base mixins.
        """
        assert isinstance(config, dict), "Invalid type: {}".format(config)
        self.config = config

        self.currency = currency
        assert isinstance(skip_lines, int)
        self.skip_lines = skip_lines
        self.last4_map = last4_map or {}
        self.debug = debug
        self.dateutil_kwds = dateutil_kwds
        self.csv_dialect = csv_dialect
        self.narration_sep = narration_sep
        self.encoding = encoding or 'utf-8'
        self.invert_sign = invert_sign
        self.categorizer = categorizer
        super().__init__(**kwds)

    def file_date(self, file):
        "Get the maximum date from the file."
        iconfig, has_header = normalize_config(
            self.config,
            file.head(encoding=self.encoding),
            self.csv_dialect,
            self.skip_lines,
        )
        if Col.DATE in iconfig:
            with open(file.name, encoding=self.encoding) as infile:
                reader = iter(csv.reader(infile, dialect=self.csv_dialect))
                for _ in range(self.skip_lines):
                    next(reader)
                if has_header:
                    next(reader)
                max_date = None
                for row in reader:
                    if not row:
                        continue
                    if row[0].startswith('#'):
                        continue
                    date_str = row[iconfig[Col.DATE]]
                    date = date_utils.parse_date(date_str, self.dateutil_kwds)
                    if max_date is None or date > max_date:
                        max_date = date
                return max_date

    def _do_extract(self, file, account, existing_entries=None):
        entries = []

        # Normalize the configuration to fetch by index.
        iconfig, has_header = normalize_config(
            self.config,
            file.head(encoding=self.encoding),
            self.csv_dialect,
            self.skip_lines,
        )

        with open(file.name, encoding=self.encoding) as infile:
            reader = iter(csv.reader(infile, dialect=self.csv_dialect))

            # Skip garbage lines
            for _ in range(self.skip_lines):
                next(reader)

            # Skip header, if one was detected.
            if has_header:
                next(reader)

            def get(row, ftype):
                try:
                    return row[iconfig[ftype]] if ftype in iconfig else None
                except IndexError:  # FIXME: this should not happen
                    return None

            # Parse all the transactions.
            first_row = last_row = None
            for index, row in enumerate(reader, 1):
                if not row:
                    continue
                if row[0].startswith('#'):
                    continue

                # If debugging, print out the rows.
                if self.debug:
                    print(row)

                if first_row is None:
                    first_row = row
                last_row = row

                # Extract the data we need from the row, based on the configuration.
                date = get(row, Col.DATE)
                txn_date = get(row, Col.TXN_DATE)
                txn_time = get(row, Col.TXN_TIME)

                payee = get(row, Col.PAYEE)
                if payee:
                    payee = payee.strip()

                fields = filter(None, [get(row, field)
                                       for field in (Col.NARRATION1,
                                                     Col.NARRATION2,
                                                     Col.NARRATION3)])
                narration = self.narration_sep.join(
                    field.strip() for field in fields).replace('\n', '; ')

                tag = get(row, Col.TAG)
                tags = {tag} if tag else data.EMPTY_SET

                link = get(row, Col.REFERENCE_ID)
                links = {link} if link else data.EMPTY_SET

                last4 = get(row, Col.LAST4)

                balance = get(row, Col.BALANCE)

                currency = get(row, Col.CURRENCY) or self.currency

                # Create a transaction
                meta = data.new_metadata(file.name, index)
                if txn_date is not None:
                    meta['date'] = date_utils.parse_date(txn_date,
                                                         self.dateutil_kwds)
                if txn_time is not None:
                    meta['time'] = str(dateutil.parser.parse(txn_time).time())
                if balance is not None:
                    meta['balance'] = Amount(self.parse_amount(balance), currency)
                if last4:
                    last4_friendly = self.last4_map.get(last4.strip())
                    meta['card'] = last4_friendly if last4_friendly else last4
                date = date_utils.parse_date(date, self.dateutil_kwds)
                txn = data.Transaction(meta, date, self.FLAG, payee, narration,
                                       tags, links, [])

                # Attach one posting to the transaction
                amount_debit, amount_credit = self.get_amounts(iconfig, row,
                                                               False, self.parse_amount)

                # Skip empty transactions
                if amount_debit is None and amount_credit is None:
                    continue

                for amount in [amount_debit, amount_credit]:
                    if amount is None:
                        continue
                    if self.invert_sign:
                        amount = -amount
                    units = Amount(amount, currency)
                    txn.postings.append(
                        data.Posting(account, units, None, None, None, None))

                # Attach the other posting(s) to the transaction.
                txn = self.call_categorizer(txn, row)

                # Add the transaction to the output list
                entries.append(txn)

        # Figure out if the file is in ascending or descending order.
        first_date = date_utils.parse_date(get(first_row, Col.DATE),
                                           self.dateutil_kwds)
        last_date = date_utils.parse_date(get(last_row, Col.DATE),
                                          self.dateutil_kwds)
        is_ascending = first_date < last_date

        # Reverse the list if the file is in descending order
        if not is_ascending:
            entries = list(reversed(entries))

        # Add a balance entry if possible. If more than one currency
        # can appear in the input, add one balance statement for each.
        if Col.BALANCE in iconfig:
            balances = set()
            for entry in reversed(entries):
                # Remove the 'balance' metadata.
                balance = entry.meta.pop('balance', None)
                if balance is None:
                    continue
                # Only add the newest entry for each currency in the file
                if balance.currency not in balances:
                    date = entry.date + datetime.timedelta(days=1)
                    meta = data.new_metadata(file.name, index)
                    entries.append(data.Balance(meta, date, account, balance, None, None))
                    balances.add(balance.currency)

        return entries

    def call_categorizer(self, txn, row):
        if not isinstance(self.categorizer, collections.abc.Callable):
            return txn

        # TODO(blais): Remove introspection here, just commit to the two
        # parameter version.
        params = signature(self.categorizer).parameters
        if len(params) < 2:
            return self.categorizer(txn)
        return self.categorizer(txn, row)

    def parse_amount(self, string):
        """The method used to create Decimal instances. You can override this."""
        return D(string)

    def get_amounts(self, iconfig, row, allow_zero_amounts, parse_amount):
        """See function get_amounts() for details.

        This method is present to allow clients to override it in order to deal
        with special cases, e.g., columns with currency symbols in them.
        """
        return get_amounts(iconfig, row, allow_zero_amounts, parse_amount)


# Deprecated. TODO(blais): Remove this eventually (on a major release).
class Importer(_CSVImporterBase, identifier.IdentifyMixin, filing.FilingMixin):
    """Importer for CSV files.

    This class implements the old ImporterProtocol interface. It is
    deprecated and will be removed eventually. Use the CSVImporter
    class instead.

    """

    def __init__(self, config, account, currency,
                 regexps=None,
                 skip_lines: int = 0,
                 last4_map: Optional[Dict] = None,
                 categorizer: Optional[Callable] = None,
                 institution: Optional[str] = None,
                 debug: bool = False,
                 csv_dialect: Union[str, csv.Dialect] = 'excel',
                 dateutil_kwds: Optional[Dict] = None,
                 narration_sep: str = '; ',
                 encoding: Optional[str] = None,
                 invert_sign: Optional[bool] = False,
                 **kwds):
        warnings.warn('beangulp.importers.csv.Importer is deprecated. '
                      'Base your importer on beangulp.importers.csvbase.Importer instead.',
                      DeprecationWarning, stacklevel=2)

        kwds.update(prepare_for_identifier(regexps, kwds.get('matchers')))
        kwds.update(prepare_for_filing(account, kwds.get('prefix', None), institution))
        super().__init__(config, account, currency,
            regexps=regexps,
            skip_lines=skip_lines,
            last4_map=last4_map,
            categorizer=categorizer,
            institution=institution,
            debug=debug,
            csv_dialect=csv_dialect,
            dateutil_kwds=dateutil_kwds,
            narration_sep=narration_sep,
            encoding=encoding,
            invert_sign=invert_sign,
            **kwds
        )

    def extract(self, file, existing_entries=None):
        account = self.file_account(file)
        return self._do_extract(file, account, existing_entries)


class CSVImporter(importer.Importer):
    """Importer for CSV files.

    This class adapts the old CSV code to implement the new redesigned
    Importer interface. The new beangulp.importers.csvbase.Importer
    may be a better base for newly developed importers.

    """

    def __init__(self, config, account, currency,
                 regexps=None,
                 skip_lines: int = 0,
                 last4_map: Optional[Dict] = None,
                 categorizer: Optional[Callable] = None,
                 institution: Optional[str] = None,
                 debug: bool = False,
                 csv_dialect: Union[str, csv.Dialect] = 'excel',
                 dateutil_kwds: Optional[Dict] = None,
                 narration_sep: str = '; ',
                 encoding: Optional[str] = None,
                 invert_sign: Optional[bool] = False,
                 **kwds):
        """See _CSVImporterBase."""

        self.base = _CSVImporterBase(config, account, currency,
                                     regexps,
                                     skip_lines,
                                     last4_map,
                                     categorizer,
                                     institution,
                                     debug,
                                     csv_dialect,
                                     dateutil_kwds,
                                     narration_sep,
                                     encoding,
                                     invert_sign)

        filing_kwds = prepare_for_filing(account, kwds.get('prefix', None), institution)
        self.filing = filing.FilingMixin(**filing_kwds)

        ident_kwds = prepare_for_identifier(regexps, kwds.get('matchers'))
        self.ident = identifier.IdentifyMixin(**ident_kwds)

    def identify(self, filepath):
        return self.ident.identify(cache.get_file(filepath))

    def account(self, filepath):
        return self.filing.file_account(cache.get_file(filepath))

    def date(self, filepath):
        return self.base.file_date(cache.get_file(filepath))

    def filename(self, filepath):
        return path.basename(utils.idify(filepath))

    def extract(self, filepath, existing=None):
        account = self.account(filepath)
        return self.base._do_extract(cache.get_file(filepath), account, existing)



__copyright__ = "Copyright (C) 2016-2017  Martin Blais"
__license__ = "GNU GPLv2"

import bisect
import datetime
import operator
import textwrap
import warnings

from typing import Callable

from beancount.core import data
from beancount.parser import printer
from beangulp import similar


# Header for the file where the extracted entries are written.
HEADER = ';; -*- mode: beancount -*-\n'

# Format for the section titles separating entries extracted from
# different documents. This is used as a format sting passing the
# document filesystem path as argument.
SECTION = '**** {}'

# Metadata field that indicates the entry is a likely duplicate.
DUPLICATE = '__duplicate__'


def extract_from_file(importer, filename, existing_entries):
    """Import entries from a document.

    Args:
      importer: The importer instance to handle the document.
      filename: Filesystem path to the document.
      existing_entries: Existing entries.

    Returns:
      The list of imported entries.
    """
    entries = importer.extract(filename, existing_entries)
    if not entries:
        return []

    # Sort the newly imported entries.
    importer.sort(entries)

    # Ensure that the entries are typed correctly.
    for entry in entries:
        data.sanity_check_types(entry)

    return entries


def sort_extracted_entries(extracted):
    """Sort the extraxted entries.

    Sort extracged entries, grouped by source document, in the order
    in which they will be used in deduplication and in which they will
    be serialized to file.

    Args:
      extracted: List of (filepath, entries, account, importer) tuples
        where entries is the list of entries extracted from the
        document at filepath by importer.

    """

    # The entries are sorted on a key composed by (max-date, account,
    # min-date, filename) where max-date and min-date are the latest
    # and earliest date appearing in the entries list.  This should
    # place entries from documents produced earlier in time at before
    # ones coming from documents produced later.
    #
    # Most imports have a balance statement at the end with a date
    # that is one day later than the reporting period (balance
    # statement are effective at the beginning of the day).  Thus
    # using the end date should be more predictable than sorting on
    # the earliest entry.
    #
    # This diagram, where the bars represents the time span covered by
    # contained entries, represent the sort order we want to obtain:
    #
    # Assets:Ali   (-----)
    # Assets:Ali   (=====--------)
    # Assets:Bob    (------------)
    # Assets:Bob             (===----)
    # Assets:Ali                 (--------------)
    # Assets:Bob                (====-----------)
    #
    # The sections marked with = represent the time spans in which
    # duplicated entries could be present.  We want entries form
    # documents produced earlier in time to take precedence over
    # entries from documents produced later in time.

    def key(element):
        filename, entries, account, importer = element
        dates = [entry.date for entry in entries]
        # Sort documents that do not contain any entry last.
        max_date = min_date = datetime.date(9999, 1, 1)
        if dates:
            max_date, min_date = max(dates), min(dates)
        return max_date, account, min_date, filename

    extracted.sort(key=key)


def find_duplicate_entries(extracted, existing):
    """Flag potentially duplicate entries.

    Args:
      extracted: List of (filepath, entries, account, importer) tuples
        where entries is the list of entries extracted from the
        document at filepath by importer.
      existing: Existing entries.

    Returns:
      A copy of the list of new entries with the potentially duplicate
      entries marked setting the "__duplicate__" metadata field to True.

    """
    # This function is kept only for backwards compatibility.
    warnings.warn('The find_duplicate_entries() function is kept only for '
                  'backwards compatibility with import scripts that explicitly '
                  'added it to the import hooks. It does not conform to the '
                  'current way of implementing deduplication and it is not to '
                  'be used.', stacklevel=2)

    ret = []
    for filepath, entries, account, importer in extracted:

        # Sort the existing entries by date: find_similar_entries()
        # uses bisection to reduce the list of existing entries to the
        # set in a narrow date interval around the date of each entry
        # in the set it is comparing against.
        existing.sort(key=operator.attrgetter('date'))

        # Find duplicates.
        pairs = similar.find_similar_entries(entries, existing)

        # We could do something smarter than throwing away the
        # information about which entry is the source of the possible
        # duplication.
        duplicates = { id(duplicate) for duplicate, source in pairs }
        marked = []
        for entry in entries:
            if id(entry) in duplicates:
                meta = entry.meta.copy()
                meta[DUPLICATE] = True
                entry = entry._replace(meta=meta)
            marked.append(entry)
        ret.append((filepath, marked, account, importer))

        # Append the current batch of extracted entries to the
        # existing entries. This allows to deduplicate entries in the
        # current extraction run.
        existing.extend(marked)

    return ret


def mark_duplicate_entries(
        entries: data.Entries,
        existing: data.Entries,
        window: datetime.timedelta,
        compare: Callable[[data.Directive, data.Directive], bool]) -> None:
    """Mark duplicate entries.

    Compare newly extracted entries to the existing entries. Only
    existing entries dated within the given time window around the
    date of the each existing entry.

    Entries that are determined to be duplicates of existing entries
    are marked setting the "__duplicate__" metadata field.

    Args:
      entries: Entries to be deduplicated.
      existing: Existing entries.
      window: Time window in which entries are compared.
      compare: Entry comparison function.

    """
    # The use of bisection to identify the entries in the existing
    # list that have dates within a given window around the date
    # of each newly extracted entry requires the existing entries
    # to be sorted by date.
    existing.sort(key=operator.attrgetter('date'))
    dates = [entry.date for entry in existing]

    def entries_date_window_iterator(date):
        lo = bisect.bisect_left(dates, date - window)
        hi = bisect.bisect_right(dates, date + window)
        for i in range(lo, hi):
            yield existing[i]

    for entry in entries:
        for target in entries_date_window_iterator(entry.date):
            if compare(entry, target):
                entry.meta[DUPLICATE] = target


def print_extracted_entries(extracted, output):
    """Print extracted entries.

    Entries marked as duplicates are printed as comments.

    Args:
      extracted: List of (filepath, entries, account, importer) tuples
        where entries is the list of entries extracted from the
        document at filepath by importer.
      output: A file object to write to. The object needs to implement
       a write() method that accepts an unicode string.

    """
    if extracted and HEADER:
        output.write(HEADER + '\n')

    for filepath, entries, account, importer in extracted:
        output.write(SECTION.format(filepath) + '\n\n')

        for entry in entries:
            duplicate = entry.meta.pop(DUPLICATE, False)
            string = printer.format_entry(entry)
            # If the entry is a duplicate, comment it out and report
            # of which other entry this is a duplicate.
            if duplicate:
                if isinstance(duplicate, type(entry)):
                    filename = duplicate.meta.get('filename')
                    lineno = duplicate.meta.get('lineno')
                    if filename and lineno:
                        output.write(f'; duplicate of {filename}:{lineno}\n')
                string = textwrap.indent(string, '; ')
            output.write(string)
            output.write('\n')

        output.write('\n')
